"""
Direct-extract-from-SAM-file
    Copyright (C) 2023  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import cupy as cp
from cupyx.scipy.spatial import distance_matrix as dm
from scipy.spatial import distance_matrix
from Bio.PDB import *
import cv2
import os
from tqdm import tqdm
import pickle as pkl


def get_coordinate(pdb_path, name):
    p = PDBParser()
    st = p.get_structure(name, pdb_path)
    ca = np.array([atom.coord for atom in st.get_atoms() if atom.name == 'CA'])
    return ca


def extract_features(mat):
    mat = cv2.normalize(mat, None, 0, 255, cv2.NORM_MINMAX).astype('uint8')
    sift = cv2.SIFT_create()
    kp, des = sift.detectAndCompute(mat, None)
    return des


def norm(x):
    x = x.T
    x1 = np.linalg.norm(x=x, ord=2, axis=0, keepdims=True)
    x = x / x1
    return x.T


def compare(des, t, data, length):
    try:
        aa = length[data[0]]
        dt = cp.asarray(data[1])
        dist = cp.dot(des, data[1].T)
        max_mat = (cp.max(dist, axis=1) - t)
        ind = max_mat > 0
        n = cp.sum(max_mat * (max_mat > 0))
        match = cp.sum(ind)
        feat_q = des.shape[0]
        feat_k = data[1].shape[0]
        if 'CELE_' in data[0]:
            name = data[0].split('_')[1]
        else:
            name = data[0]
        return f"{name},{match},{n},{feat_q},{feat_k},{aa}\n"
    except:
        return f"{0},{0},{0},{0},{0},{0},{0}\n"



"""
Direct-extract-from-SAM-file
    Copyright (C) 2023
    2q3

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from tool_kit import *
import cupy as cp
from cupyx.scipy.spatial import distance_matrix
from scipy.spatial import distance_matrix as dm
import numpy as np
from functools import partial
from tqdm.contrib.concurrent import process_map
from tqdm import tqdm





class ADAMS_match():
    def __init__(self, ref, threshold=0.95):
        self.ref = ref
        self.features = cp.array([])
        self.distance_matrix = np.array([])
        self.threshold = threshold
        self.extract_features()

    def extract_features(self):
        ca = get_coordinate(self.ref, 'X')
        dist = distance_matrix(ca, ca)
        features = extract_features(dist)
        features = norm(features)
        self.distance_matrix = dist
        self.features = cp.array(features)

    def match(self, db_folder, out):
        mempool = cp.get_default_memory_pool()
        mempool.set_limit(size=6 * 1024 ** 3)
        files = os.listdir(db_folder)
        config_path = os.path.join(db_folder, [n for n in files if n[-6:] == 'config'][0])
        with open(config_path, 'r') as f:
            database_list = f.readlines()
        print('{} PARTS OF THE DATABASE WAITING FOR COMPARISON'.format(len(database_list)))
        result = []
        database_list = [n[:-1] for n in database_list]
        for i in database_list:
            with open(os.path.join(db_folder, i), 'rb') as f:
                database = pkl.load(f)
            with open(os.path.join(db_folder, i[:-4] + '_len.pkl'), 'rb') as f:
                length = pkl.load(f)
            data = list(database.items())
            for i in tqdm(data):
                result.append(compare(self.features, self.threshold, i, length))
            mempool.free_all_blocks()
            del database, data
        with open(out, 'w+') as f:
            f.write('protein,match,score,feat_q,feat_k,aa_length\n')
            for i in result:
                if i[0] != '0':
                    try:
                        f.write(i)
                    except:
                        continue



"""
Direct-extract-from-SAM-file
    Copyright (C) 2023  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from tool_kit import *
import os
from scipy.spatial import distance_matrix
import cupy as cp
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map


def get_feature_list(pdb_folder, file):
    error = []
    try:
        ca = get_coordinate(os.path.join(pdb_folder, file), 'X')
        name = file[:-4]
        dist = distance_matrix(ca, ca)
        dist = extract_features(dist)
        dist = norm(dist)
        length = len(ca)
        return (name, dist, length)
    except:
        error.append('1')
        return ('error', 0, 0)


class DatabaseMaker():
    def __init__(self, device=0, chunk_size=5000, process=40):
        self.device = device
        self.chunk_size = chunk_size
        self.process = process

    def make(self, pdb_folder, out_perfix):
        mempool = cp.get_default_memory_pool()
        files = os.listdir(pdb_folder)
        chunks = len(files) // self.chunk_size
        print('total_chunks = {}'.format(chunks + 1))
        feature_p = partial(get_feature_list, pdb_folder)
        db_list = []
        for i in range(chunks + 1):
            feature_dict = {}
            length_dict = {}
            file = files[5000 * i:5000 * i + 5000]
            result = process_map(feature_p, file, max_workers=self.process, chunksize=1)
            for j in range(len(result)):
                feature_dict[result[j][0]] = cp.asarray(result[j][1])
                length_dict[result[j][0]] = cp.asarray(result[j][2])
            db_list.append(out_perfix + '{}.pkl'.format(i))
            with open(out_perfix + '{}.pkl'.format(i), 'wb') as f:
                pkl.dump(feature_dict, f)
            with open(out_perfix + '{}_len.pkl'.format(i), 'wb') as f:
                pkl.dump(length_dict, f)
            mempool.free_all_blocks()
        with open(out_perfix + '.config', 'w+') as f:
            for i in db_list:
                f.write(i + '\n')


from setuptools import find_packages
from distutils.core import setup

install = ['numpy >= 1.17.2', 'opencv-python >= 4.7.0.12', 'scipy == 1.11.2', 'tqdm == 4.66.1', 'biopython == 1.81']


with open("Readme.md", "r", encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='adams',
    version='0.0.1',
    description='this is a program for fast protein structure search',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Guo_Zhengyang',
    author_email='guozy23@mails.tsinghua.edu.cn',
    install_requires=install,
    packages=find_packages(),
    license='GNU-GPL 3.0',
    classifiers=['Programming Language :: Python :: 3', 'Development Status :: 4 - Beta', ],
    url='https://github.com/young55775/ADAMS-developing',
    include_package_data=True
)
